/*
md.c - MD-PRO flash card support for uCON64

written by 2003 dbjh
           2003 NoisyB (noisyb@gmx.net)
Based on Delphi source code by ToToTEK Multi Media. Information in that source
code has been used with permission. However, ToToTEK Multi Media explicitly
stated that the information in that source code may be freely distributed.


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef  HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "misc.h"
#include "quick_io.h"
#include "ucon64.h"
#include "ucon64_dat.h"
#include "ucon64_misc.h"
#include "md.h"
#include "misc_par.h"


const st_usage_t md_usage[] =
  {
    {NULL, NULL, "MD-PRO flash card programmer"},
    {NULL, NULL, "2003 ToToTEK Multi Media http://www.tototek.com"},
#ifdef PARALLEL
    {"xmd", NULL, "send/receive ROM to/from MD-PRO flash card programmer\n"
                  OPTION_LONG_S "port=PORT\n"
                  "receives automatically (32 Mbits) when ROM does not exist"},
    {"xmds", NULL, "send/receive SRAM to/from MD-PRO flash card programmer\n"
                  OPTION_LONG_S "port=PORT\n"
                  "receives automatically when SRAM does not exist"},
#endif // PARALLEL
    {NULL, NULL, NULL}
  };


#ifdef PARALLEL

static void init_io (unsigned int port);
static void deinit_io (void);
static void set_data_read (void);
static void set_data_write (void);
static void epp_set_ai (unsigned char ai);
static void set_ai_data (unsigned char ai, unsigned char data);
static void set_ai (unsigned char ai);
static void init_port (void);
static void deinit_port (void);
static void end_port (void);
static void romCS_on (void);
static void romCS_off (void);
static void ramCS_on (void);
static void ramCS_off (void);
static void set_long_addr_read (int addr);
static void set_long_addr_write (int addr);
static void writeMEMb (int addr, unsigned char b);
static void EEP_reset (void);
static unsigned char getID (unsigned char addr);
static unsigned short int getIDword (void);
static void erase_block (int addr);
#if 0 // not used
static void readstatus (void);
#endif
static void writeEEPDataB (int addr, unsigned char b);
static void writeIntelEEPDataB (int addr, unsigned char b);
static void writeEEPDataPage (int addr, unsigned char *buf);
static void readpagerambuf (int addr, unsigned short int *buf);
static void readrambuf (int addr, unsigned char *buf);
static void writeRAMDataB (int addr, unsigned char b);
#if 0 // not used
static void writeRAMDataPage (int addr, unsigned char *buf);
#endif
static void writeRAMDBDataPage (int addr, unsigned char *buf);

static void write_block_by_byte (unsigned char *buf, int *addr);
static void write_block_by_page (unsigned char *buf, int *addr);

//static void EEPVerifyFileByPage (char *fn, int beginaddr, unsigned char MDtype);

static void read_buf (int addr, unsigned char *buf);
static void read_bufb (int addr, unsigned char *buf);

static void write_sram_by_byte (unsigned char *buf, int *addr);
static void write_sram_by_page (unsigned char *buf, int *addr);

static unsigned short int md_idw;
static short int port_8, port_9, port_a, port_b, port_c;


void
init_io (unsigned int port)
{
#if     (defined __unix__ || defined __BEOS__) && !defined __MSDOS__
  init_conio ();
#endif

  port_8 = port;                                // original code used 0x378 for port_8
  port_9 = port + 1;
  port_a = port + 2;
  port_b = port + 3;
  port_c = port + 4;

  misc_parport_print_info ();

  init_port ();

  (void) readpagerambuf;                        // warning remover
  (void) readrambuf;                            // warning remover
  (void) write_sram_by_page;                    // warning remover
}


void
deinit_io (void)
{
#if     (defined __unix__ || defined __BEOS__) && !defined __MSDOS__
  deinit_conio ();
#endif

  end_port ();
  deinit_port ();
}


// ported ToToTEK code starts here
void
set_data_read (void)
{
  outportb (port_a, 0);                 // nastb=1,nib_sel=0,ndstb=1,nwrite=1
}


void
set_data_write (void)
{
  outportb (port_a, 1);                 // nastb=1,nib_sel=0,ndstb=1,nwrite=0
}


void
epp_set_ai (unsigned char ai)
{
  outportb (port_a, 1);                 // set_data_write ()
  outportb (port_b, ai);
}


void
set_ai_data (unsigned char ai, unsigned char data)
{
  epp_set_ai (ai);
  outportb (port_a, 1);                 // set_data_write ()
  outportb (port_c, data);
}


void
set_ai (unsigned char ai)
{
  set_data_write ();
  outportb (port_b, ai);
}


void
init_port (void)
{
  outportb ((unsigned short) (port_8 + 0x402),
            (unsigned char) ((inportb ((unsigned short) (port_8 + 0x402)) & 0x1f) | 0x80));
#ifndef PPDEV // probably #if 0, but first test if this works with ppdev - dbjh
  outportb (port_9, 1);                         // clear EPP time flag
#endif
  set_ai_data (6, 0);                           // rst=0, wei=0(dis.), rdi=0(dis.)
  set_ai_data (6, 0x84);                        // rst=1, wei=0(dis.), rdi=0(dis.)
}


void
deinit_port (void)
{
  outportb (port_a, 1);
  outportb (port_b, 0);
  outportb (port_a, 4);                         // port normal now
}


void
end_port (void)
{
  set_ai_data (6, 0);                           // rst=0, wei=0(dis.), rdi=0(dis.)
  outportb (port_a, 4);                         // set_normal ninit=1, nwrite=1
}


void
romCS_on (void)
{
  set_ai_data (6, 0x84);
}


void
romCS_off (void)
{
  set_ai_data (6, 0x80);
}


void
ramCS_on (void)
{
  set_ai_data (6, 0x88);
}


void
ramCS_off (void)
{
  set_ai_data (6, 0x80);
}


void
set_long_addr_read (int addr)
{
  set_ai_data (0, (unsigned char) (addr & 0xff)); // a[7..0]
  addr >>= 8;
  set_ai_data (1, (unsigned char) (addr & 0xff)); // a[15..8]
  addr >>= 8;
  set_ai_data (2, (unsigned char) (addr & 0xff)); // a[23..16]
  set_ai (3);
  set_data_read ();                             // ninit=0, nwrite=1
}


void
set_long_addr_write (int addr)
{
  set_ai_data (0, (unsigned char) (addr & 0xff)); // a[7..0]
  addr >>= 8;
  set_ai_data (1, (unsigned char) (addr & 0xff)); // a[15..8]
  addr >>= 8;
  set_ai_data (2, (unsigned char) (addr & 0xff)); // a[23..16]
  set_ai (3);
}


void
writeMEMb (int addr, unsigned char b)
{
  set_ai_data (0, (unsigned char) (addr & 0xff)); // a[7..0]
  addr >>= 8;
  set_ai_data (1, (unsigned char) (addr & 0xff)); // a[15..8]
  addr >>= 8;
  set_ai_data (2, (unsigned char) (addr & 0xff)); // a[23..16]
  set_ai (3);
  outportb (port_c, b);
}


void
EEP_reset (void)
{
  romCS_on ();
  writeMEMb (0x400000, 0xff);                   // reset EEP Chip2
  writeMEMb (0, 0xff);                          // reset EEP Chip1
  writeMEMb (0x600000, 0xff);                   // reset EEP Chip2
  writeMEMb (0x200000, 0xff);                   // reset EEP Chip1
  romCS_off ();
}


unsigned char
getID (unsigned char addr)
{
  unsigned char c;

  romCS_on ();
  set_ai_data (0, addr);                        // a[7..0] = 0
  set_ai (3);
  outportb (port_c, 0x90);                      // write_data ()

  set_ai (3);
  set_data_read ();                             // ninit=0, nwrite=1
  c = inportb (port_c);                         // read_data ()
  romCS_off ();

  return c;
}


unsigned short int
getIDword (void)
{
  unsigned short int w;

//  EEP_reset ();
  w = getID (0);                                // Msg
  w <<= 8;
  w |= getID (2);                               // ID
  return w;
}


/*=================== SHARP function */
void
erase_block (int addr)
{
  unsigned char d;

  romCS_on ();
  writeMEMb (addr, 0x50);
  romCS_off ();

  romCS_on ();
  writeMEMb (addr, 0x20);
  outportb (port_c, 0xd0);
  romCS_off ();

  romCS_on ();
  set_ai (3);
  set_data_read ();                             // set read mode
  d = inportb (port_c);
  while (d < 0x80)
    d = inportb (port_c);
  romCS_off ();
}


#if 0
void
readstatus (void)
{
  unsigned char d;

  // send command 0xe8
  romCS_on ();
  set_ai (3);
  set_data_read ();
  d = inportb (port_c);
  while (d < 0x80)
    d = inportb (port_c);
  romCS_off ();
}
#endif


void
writeEEPDataB (int addr, unsigned char b)
{
  romCS_on ();
  writeMEMb (addr, 0x40);
  outportb (port_c, b);
  romCS_off ();
}


void
writeIntelEEPDataB (int addr, unsigned char b)
{
  unsigned char d;

  romCS_on ();
  writeMEMb (addr, 0x40);
  outportb (port_c, b);

  set_data_read ();
  d = inportb (port_c);
  while (d < 0x80)
    d = inportb (port_c);
  romCS_off ();
}


void
writeEEPDataPage (int addr, unsigned char *buf)
{
  unsigned char d, i;

  // send command 0xe8
  romCS_on ();
  writeMEMb (addr, 0xe8);

  // read status
  set_ai (3);
  set_data_read ();
  d = inportb (port_c);
  while (d < 0x80)
    d = inportb (port_c);

  set_ai (3);
  outportb (port_c, 0x1f);
  set_ai_data (6, 0x94);
  set_long_addr_write (addr);
  for (i = 0; i <= 0x1f; i++)
    outportb (port_c, buf[(addr + i) & 0x3fff]);
  set_ai_data (6, 0x84);
  set_ai (3);
  outportb (port_c, 0xd0);

  // read status
  set_ai (3);
  set_data_read ();
  d = inportb (port_c);
  while (d < 0x80)
    d = inportb (port_c);
  romCS_off ();
}


void
readpagerambuf (int addr, unsigned short int *buf)
{
  int count;

  set_long_addr_read (addr);
  for (count = 0; count < 0x80; count++)
    buf[count] = inportw (port_c);              // read_dataw (); data is doubled
}                                               //  => no problems with endianess


void
readrambuf (int addr, unsigned char *buf)
{
  int count;

  ramCS_on ();
  for (count = 0; count < 0x100; count++)
    {
      set_long_addr_read (addr + count);
      buf[count] = inportb (port_c);            // read_data ()
    }
  ramCS_off ();
}


void
writeRAMDataB (int addr, unsigned char b)
{
  ramCS_on ();
  writeMEMb (addr, b);
  ramCS_off ();
}


#if 0
void
writeRAMDataPage (int addr, unsigned char *buf)
{
  unsigned char i;

  ramCS_on ();
  set_ai_data (6, 0x98);
  set_long_addr_write (addr);
  for (i = 0; i < 0x100; i++)
    outportb (port_c, buf[(addr + i) & 0x3fff]);
  ramCS_off ();
}
#endif


void
writeRAMDBDataPage (int addr, unsigned char *buf)
{
  unsigned char i;

  ramCS_on ();
  set_ai_data (6, 0x98);
  set_long_addr_write (addr * 2);
  for (i = 0; i < 0x80; i++)
    {
      outportb (port_c, buf[(addr + i) & 0x3fff]);
      outportb (port_c, buf[(addr + i) & 0x3fff]);
    }
  ramCS_off ();
}


void
write_block_by_byte (unsigned char *buf, int *addr)
{
  int x;

  for (x = 0; x < 0x4000; x++)
    {
      if (md_idw == 0xb0d0)
        writeEEPDataB (*addr, buf[*addr & 0x3fff]);
      else if (md_idw == 0x8917)
        writeIntelEEPDataB (*addr, buf[*addr & 0x3fff]);
      (*addr)++;
    }
}


void
write_block_by_page (unsigned char *buf, int *addr)
{
  int x;

  for (x = 0; x < 0x200; x++)
    {
      writeEEPDataPage (*addr, buf);
      (*addr) += 0x20;
    }
}


void
read_bufb (int addr, unsigned char *buf)
{
  int count;

  romCS_on ();
  for (count = 0; count < 0x100; count++)
    {
      set_long_addr_read (addr + count);
      buf[count] = inportb (port_c);            // read_data ()
    }
  romCS_off ();
}


void
read_buf (int addr, unsigned char *buf)
{
  int count;

  set_long_addr_read (addr);
  for (count = 0; count < 0x80; count++)
#ifdef  WORDS_BIGENDIAN
    ((unsigned short int *) buf)[count] = bswap_16 (inportw (port_c)); // read_dataw ()
#else
    ((unsigned short int *) buf)[count] = inportw (port_c); // read_dataw ()
#endif
}


void
write_sram_by_byte (unsigned char *buf, int *addr)
{
  int x, i = *addr & 0x3fff;

  for (x = 0; x < 0x4000; x++, i = (i + 1) & 0x3fff)
    {
      writeRAMDataB (*addr, buf[i]);
      (*addr)++;
      // Send the same byte again => SRAM files needn't store redundant data
      writeRAMDataB (*addr, buf[i]);
      (*addr)++;
    }
}


void
write_sram_by_page (unsigned char *buf, int *addr)
{
  int x;

  for (x = 0; x < 0x80; x++)
    {
      writeRAMDBDataPage (*addr, buf);
      (*addr) += 0x80;
    }
}
// ported ToToTEK code ends here


int
md_read_rom (const char *filename, unsigned int parport, int size)
{
  FILE *file;
  unsigned char buffer[0x100];
  int blocksleft, address = 0, bytesreceived = 0;
  time_t starttime;
  void (*read_block) (int, unsigned char *) = read_bufb; // read_buf

  init_io (parport);

  if ((file = fopen (filename, "wb")) == NULL)
    {
      fprintf (stderr, ucon64_msg[OPEN_WRITE_ERROR], filename);
      exit (1);
    }

  blocksleft = size >> 8;
  EEP_reset ();
  romCS_on ();
  if (read_block == read_buf)
    set_ai_data (6, 0x94);          // rst=1, wei=0(dis.), rdi=0(dis.), inc mode, rom_CS
  starttime = time (NULL);
  while (blocksleft-- > 0)
    {
      read_block (address, buffer);             // 0x100 bytes read
      if (read_block == read_bufb)
        mem_swap_b (buffer, 0x100);
      fwrite (buffer, 1, 0x100, file);
      address += 0x100;
      bytesreceived += 0x100;
      if ((address & 0x3fff) == 0)
        ucon64_gauge (starttime, bytesreceived, size);
    }
  // original code doesn't call romCS_off() when byte-size function is used
  if (read_block == read_buf)
    romCS_off ();

  fclose (file);
  deinit_io ();

  return 0;
}


int
md_write_rom (const char *filename, unsigned int parport)
{
  FILE *file;
  unsigned char *buffer;
  int fsize, address = 0, bytessend = 0;
  time_t starttime;
  void (*write_block) (unsigned char *buf, int *addr) = write_block_by_page; // write_block_by_byte
  (void) write_block_by_byte;

  init_io (parport);

  if ((file = fopen (filename, "rb")) == NULL)
    {
      fprintf (stderr, ucon64_msg[OPEN_READ_ERROR], filename);
      exit (1);
    }
  if ((buffer = (unsigned char *) malloc (0x4000)) == NULL)
    {
      fprintf (stderr, ucon64_msg[FILE_BUFFER_ERROR], 0x4000);
      exit (1);
    }

  fsize = q_fsize (filename);
  printf ("Send: %d Bytes (%.4f Mb)\n", fsize, (float) fsize / MBIT);

  starttime = time (NULL);
  EEP_reset ();
  md_idw = getIDword ();
  if ((md_idw == 0xb0d0) || (md_idw == 0x8917)) // sharp 32M, Intel 64J3
    {
      EEP_reset ();
      while (fread (buffer, 0x4000, 1, file))   // note order of size and number args
        {
          mem_swap_b (buffer, 0x4000);
          if ((((address & 0xffff) == 0) && (md_idw == 0xb0d0)) ||
              (((address & 0x1ffff) == 0) && (md_idw == 0x8917)))
            erase_block (address);
          write_block (buffer, &address);
          bytessend += 0x4000;
          ucon64_gauge (starttime, bytessend, fsize);
        }
    }
  else
    fprintf (stderr, "ERROR: MD-PRO flash card programmer not detected\n");

  free (buffer);
  fclose (file);
  deinit_io ();

  return 0;
}


int
md_read_sram (const char *filename, unsigned int parport)
/*
  The MD-PRO has 256 kB of SRAM. However, the SRAM dumps of all games that have
  been tested had each byte doubled. In order to make it possible to easily
  obtain the SRAM data for use in an emulator, or to send an emulator SRAM file
  to the MD-PRO, we remove the redundant data when receiving/dumping and double
  the data when sending.
  It could be that this approach causes trouble for some games. However, when
  looking at ToToTEK's own code in writeRAMDBDataPage() this seems unlikely
  (data is doubled in that function). Note that write_sram_by_byte() is a
  function written by us, and so does the doubling of data, but it doesn't mean
  it should work for all games.
*/
#if 1
{
  FILE *file;
  unsigned char buffer[0x100];
  int blocksleft, address = 0, bytesreceived = 0, size = 128 * 1024, i;
  time_t starttime;

  init_io (parport);

  if ((file = fopen (filename, "wb")) == NULL)
    {
      fprintf (stderr, ucon64_msg[OPEN_WRITE_ERROR], filename);
      exit (1);
    }

//  set_ai_data (6, 0x94);                // rst=1, wei=0(dis.), rdi=0(dis.), inc mode, rom_CS

  blocksleft = size >> 7;
  starttime = time (NULL);
  while (blocksleft-- > 0)
    {
      readrambuf (address, buffer);             // 0x100 bytes read
      for (i = 0; i < 0x80; i++)
        buffer[i] = buffer[2 * i];              // data is doubled => no problems with endianess
      fwrite (buffer, 1, 0x80, file);
      address += 0x100;
      bytesreceived += 0x80;
      if ((address & 0x3fff) == 0)
        ucon64_gauge (starttime, bytesreceived, size);
    }

  fclose (file);
  deinit_io ();

  return 0;
}
#else // this code doesn't seem to work
{
  FILE *file;
  unsigned char buffer[0x100];
  int blocksleft, address = 0 /* startaddress * 2 */, bytesreceived = 0,
      size = 128 * 1024, i;
  time_t starttime;

  init_io (parport);

  if ((file = fopen (filename, "wb")) == NULL)
    {
      fprintf (stderr, ucon64_msg[OPEN_WRITE_ERROR], filename);
      exit (1);
    }

  ramCS_on ();
  set_ai_data (6, 0x98);                // rst=1, wei=0(dis.), rdi=0(dis.), inc mode, rom_CS
  blocksleft = size >> 7;
  starttime = time (NULL);
  while (blocksleft-- > 0)
    {
      readpagerambuf (address, (unsigned short int *) buffer); // 0x100 bytes read
      for (i = 0; i < 0x80; i++)
        buffer[i] = buffer[2 * i];              // data is doubled => no problems with endianess
      fwrite (buffer, 1, 0x80, file);
      address += 0x100;
      bytesreceived += 0x80;
      if ((address & 0x3fff) == 0)
        ucon64_gauge (starttime, bytesreceived, size);
    }
  ramCS_off ();

  fclose (file);
  deinit_io ();

  return 0;
}
#endif


int
md_write_sram (const char *filename, unsigned int parport)
{
  FILE *file;
  unsigned char *buffer;
  int size, bytessend = 0, address = 0;
  time_t starttime;
  void (*write_sram) (unsigned char *, int *) = write_sram_by_byte; // write_sram_by_page

  init_io (parport);

  if ((file = fopen (filename, "rb")) == NULL)
    {
      fprintf (stderr, ucon64_msg[OPEN_READ_ERROR], filename);
      exit (1);
    }
  if ((buffer = (unsigned char *) malloc (0x4000)) == NULL)
    {
      fprintf (stderr, ucon64_msg[FILE_BUFFER_ERROR], 0x4000);
      exit (1);
    }

  size = q_fsize (filename);
  printf ("Send: %d Bytes\n", size);

  starttime = time (NULL);
  while (fread (buffer, 0x4000, 1, file))       // note order of size and number args
    {
      write_sram (buffer, &address);
      bytessend += 0x4000;
      ucon64_gauge (starttime, bytessend, size);
    }

  free (buffer);
  fclose (file);
  deinit_io ();

  return 0;
}

#endif // PARALLEL
