/*
misc_par.c - miscellaneous parallel port functions

written by 1999 - 2003 NoisyB (noisyb@gmx.net)
           2001 - 2004 dbjh
           2001        Caz
           2002 - 2003 Jan-Erik Karlsson (Amiga)


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef  HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef  PARALLEL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef  HAVE_UNISTD_H
#include <unistd.h>                             // ioperm() (libc5)
#endif

#ifdef  PPDEV                                   // ppdev is a Linux parallel
#include <fcntl.h>                              //  port device driver
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/ppdev.h>
#include <linux/parport.h>
#elif   defined __linux__ && defined __GLIBC__  // PPDEV
#ifdef  HAVE_SYS_IO_H                           // necessary for some Linux/PPC configs
#include <sys/io.h>                             // ioperm() (glibc), in{b, w}(), out{b, w}()
#else
#error No sys/io.h; configure with --disable-parallel
#endif
#elif   defined __BEOS__ || defined __FreeBSD__ // __linux__ && __GLIBC__
#include <fcntl.h>
#elif   defined AMIGA                           // __BEOS__ || __FreeBSD__
#include <fcntl.h>
#include <exec/types.h>
#include <exec/io.h>
#include <exec/ports.h>
#include <dos/dos.h>
#include <dos/var.h>
#include <devices/parallel.h>
#elif   defined _WIN32                          // AMIGA
#include <windows.h>
#include <conio.h>                              // inp{w}() & outp{w}()
#include "dlopen.h"
#elif   defined __CYGWIN__                      // _WIN32
#include <windows.h>                            // definition of WINAPI
#undef  _WIN32
#include "dlopen.h"
#endif

#include "ucon64.h"
#include "misc_par.h"
#include "misc.h"


#if     defined PPDEV || defined __BEOS__ || defined __FreeBSD__
static int misc_io_fd;
#ifdef  PPDEV
static enum { FORWARD = 0, REVERSE } misc_io_direction;
static int misc_io_mode;
#endif
#endif

#ifdef  __BEOS__
typedef struct st_ioport
{
  unsigned int port;
  unsigned char data8;
  unsigned short data16;
} st_ioport_t;
#endif

#ifdef  AMIGA
static struct IOStdReq *misc_io_req;
static struct MsgPort *misc_parport;
#endif


#ifdef  __i386__                                // GCC && x86
inline static unsigned char
i386_input_byte (unsigned short port)
{
  unsigned char byte;
  __asm__ __volatile__
  ("inb %1, %0"
    : "=a" (byte)
    : "d" (port)
  );
  return byte;
}


inline static unsigned short
i386_input_word (unsigned short port)
{
  unsigned short word;
  __asm__ __volatile__
  ("inw %1, %0"
    : "=a" (word)
    : "d" (port)
  );
  return word;
}


inline static void
i386_output_byte (unsigned short port, unsigned char byte)
{
  __asm__ __volatile__
  ("outb %1, %0"
    :
    : "d" (port), "a" (byte)
  );
}


inline static void
i386_output_word (unsigned short port, unsigned short word)
{
  __asm__ __volatile__
  ("outw %1, %0"
    :
    : "d" (port), "a" (word)
  );
}
#endif // __i386__


#if     defined _WIN32 || defined __CYGWIN__
void *io_driver;

// inpout32.dll only has I/O functions for byte-sized I/O
unsigned char (__stdcall *Inp32) (unsigned short) = NULL;
void (__stdcall *Outp32) (unsigned short, unsigned char) = NULL;

unsigned char inpout32_input_byte (unsigned short port) { return Inp32 (port); }
void inpout32_output_byte (unsigned short port, unsigned char byte) { Outp32 (port, byte); }

// io.dll has more functions then the ones we refer to here, but we don't need them
char (WINAPI *PortIn) (short int) = NULL;
short int (WINAPI *PortWordIn) (short int) = NULL;
void (WINAPI *PortOut) (short int, char) = NULL;
void (WINAPI *PortWordOut) (short int, short int) = NULL;
short int (WINAPI *IsDriverInstalled) () = NULL;

unsigned char io_input_byte (unsigned short port) { return PortIn (port); }
unsigned short io_input_word (unsigned short port) { return PortWordIn (port); }
void io_output_byte (unsigned short port, unsigned char byte) { PortOut (port, byte); }
void io_output_word (unsigned short port, unsigned short word) { PortWordOut (port, word); }

#if     defined __CYGWIN__ || defined __MINGW32__
// default to functions which are always available (but which generate an
//  exception under Windows NT/2000/XP without a "driver" such as UserPort)
unsigned char (*input_byte) (unsigned short) = i386_input_byte;
unsigned short (*input_word) (unsigned short) = i386_input_word;
void (*output_byte) (unsigned short, unsigned char) = i386_output_byte;
void (*output_word) (unsigned short, unsigned short) = i386_output_word;

#elif   defined _WIN32
// The following four functions are needed because inp{w} and outp{w} seem to be macros
unsigned char inp_func (unsigned short port) { return (unsigned char) inp (port); }
unsigned short inpw_func (unsigned short port) { return inpw (port); }
void outp_func (unsigned short port, unsigned char byte) { outp (port, byte); }
void outpw_func (unsigned short port, unsigned short word) { outpw (port, word); }

// default to functions which are always available (but which generate an
//  exception under Windows NT/2000/XP without a "driver" such as UserPort)
unsigned char (*input_byte) (unsigned short) = inp_func;
unsigned short (*input_word) (unsigned short) = inpw_func;
void (*output_byte) (unsigned short, unsigned char) = outp_func;
void (*output_word) (unsigned short, unsigned short) = outpw_func;

#endif
#endif // defined _WIN32 || defined __CYGWIN__


#if     defined PPDEV || defined __BEOS__ || defined __FreeBSD__ || defined AMIGA
void
close_io_port (void)
{
#ifdef  AMIGA
  CloseDevice ((struct IORequest *) misc_io_req);
  DeleteExtIO ((struct IOExtPar *) misc_io_req);
  DeletePort (misc_parport);
  misc_io_req = NULL;
#else
#ifdef  PPDEV
  ioctl (misc_io_fd, PPRELEASE);
#endif
  close (misc_io_fd);
#endif
}
#endif


unsigned char
inportb (unsigned short port)
{
#ifdef  PPDEV
  int ppreg = port - ucon64.parport;
  unsigned char byte;

  switch (ppreg)
    {
    case 0:                                     // data
      if (misc_io_direction == FORWARD)         // dir is forward?
        {
          misc_io_direction = REVERSE;          // change it to reverse
          ioctl (misc_io_fd, PPDATADIR, &misc_io_direction);
        }
      ioctl (misc_io_fd, PPRDATA, &byte);
      break;
    case 1:                                     // status
      ioctl (misc_io_fd, PPRSTATUS, &byte);
      break;
    case 2:                                     // control
      ioctl (misc_io_fd, PPRCONTROL, &byte);
      break;
    case 3:                                     // EPP/ECP address
      if (!(misc_io_mode & IEEE1284_ADDR))      // IEEE1284_DATA is 0!
        {
          misc_io_mode |= IEEE1284_ADDR;
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      read (misc_io_fd, &byte, 1);
      break;
    case 4:                                     // EPP/ECP data
      if (misc_io_mode & IEEE1284_ADDR)
        {
          misc_io_mode &= ~IEEE1284_ADDR;       // IEEE1284_DATA is 0
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      read (misc_io_fd, &byte, 1);
      break;
    case 0x402:                                 // ECP register
      printf ("WARNING: Ignored read from ECP register, returning 0\n");
      byte = 0;
      break;
    default:
      fprintf (stderr,
               "ERROR: inportb() tried to read from an unsupported port (0x%x)\n",
               port);
      exit (1);
    }
  return byte;
#elif   defined __BEOS__
  st_ioport_t temp;

  temp.port = port;
  ioctl (misc_io_fd, 'r', &temp, 0);

  return temp.data8;
#elif   defined AMIGA
  ULONG wait_mask;
  (void) port;                                  // warning remover

  misc_io_req->io_Length = 1;
  misc_io_req->io_Command = CMD_READ;

/*
  SendIO ((struct IORequest *) misc_io_req);

  wait_mask = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F | 1L << misc_parport->mp_SigBit;
  if (Wait (wait_mask) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F))
    AbortIO ((struct IORequest *) misc_io_req);
  WaitIO ((struct IORequest *) misc_io_req);
*/

  /*
    The difference between using SendIO() and DoIO(), is that DoIO() handles
    messages etc. by itself but it will not return until the IO is done.

    Probably have to do more error handling here :-)

    Can one CTRL-C a DoIO() request? (Or for that matter a SendIO().)
  */

  if (DoIO ((struct IORequest *) misc_io_req))
    {
      fprintf (stderr, "ERROR: Could not communicate with parallel port:\n"
                       "       %s\n", misc_io_req->IOPar.io_Error);
      exit (1);
    }

  return (unsigned char) misc_io_req->io_Data;
#elif   defined _WIN32 || defined __CYGWIN__
  return input_byte (port);
#elif   defined __i386__
  return i386_input_byte (port);
#elif   defined HAVE_SYS_IO_H
  return inb (port);
#endif
}


unsigned short
inportw (unsigned short port)
{
#ifdef  PPDEV
  int ppreg = port - ucon64.parport;
  unsigned char buf[2];

  switch (ppreg)
    {
    case 3:                                     // EPP/ECP address
      if (!(misc_io_mode & IEEE1284_ADDR))      // IEEE1284_DATA is 0!
        {
          misc_io_mode |= IEEE1284_ADDR;
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      read (misc_io_fd, buf, 2);
      break;
    case 4:                                     // EPP/ECP data
      if (misc_io_mode & IEEE1284_ADDR)
        {
          misc_io_mode &= ~IEEE1284_ADDR;       // IEEE1284_DATA is 0
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      read (misc_io_fd, buf, 2);
      break;
    // the data, status, control and ECP registers should only be accessed in "8-bit mode"
    default:
      fprintf (stderr,
               "ERROR: inportw() tried to read from an unsupported port (0x%x)\n",
               port);
      exit (1);
    }
  return buf[0] | buf[1] << 8;                  // words are read in little endian format
#elif   defined __BEOS__
  st_ioport_t temp;

  temp.port = port;
  ioctl (misc_io_fd, 'r16', &temp, 0);

  return temp.data16;
#elif   defined AMIGA
  ULONG wait_mask;
  (void) port;                                  // warning remover

  misc_io_req->io_Length = 2;
  misc_io_req->io_Command = CMD_READ;

/*
  SendIO ((struct IORequest *) misc_io_req);

  wait_mask = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F | 1L << misc_parport->mp_SigBit;
  if (Wait (wait_mask) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F))
    AbortIO ((struct IORequest *) misc_io_req);
  WaitIO ((struct IORequest *) misc_io_req);
*/

  if (DoIO ((struct IORequest *) misc_io_req))
    {
      fprintf (stderr, "ERROR: Could not communicate with parallel port:\n"
                       "       %s\n", misc_io_req->IOPar.io_Error);
      exit (1);
    }

  return (unsigned short) misc_io_req->io_Data;
#elif   defined _WIN32 || defined __CYGWIN__
  return input_word (port);
#elif   defined __i386__
  return i386_input_word (port);
#elif   defined HAVE_SYS_IO_H
  return inw (port);
#endif
}


void
outportb (unsigned short port, unsigned char byte)
{
#ifdef  PPDEV
  int ppreg = port - ucon64.parport;

  switch (ppreg)
    {
    case 0:                                     // data
      if (misc_io_direction == REVERSE)         // dir is reverse?
        {
          misc_io_direction = FORWARD;          // change it to forward
          ioctl (misc_io_fd, PPDATADIR, &misc_io_direction);
        }
      ioctl (misc_io_fd, PPWDATA, &byte);
      break;
    case 2:                                     // control
      ioctl (misc_io_fd, PPWCONTROL, &byte);
      break;
    case 3:                                     // EPP/ECP address
      if (!(misc_io_mode & IEEE1284_ADDR))      // IEEE1284_DATA is 0!
        {
          misc_io_mode |= IEEE1284_ADDR;
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      write (misc_io_fd, &byte, 1);
      break;
    case 4:                                     // EPP/ECP data
      if (misc_io_mode & IEEE1284_ADDR)
        {
          misc_io_mode &= ~IEEE1284_ADDR;       // IEEE1284_DATA is 0
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      write (misc_io_fd, &byte, 1);
      break;
    case 0x402:                                 // ECP register
      printf ("WARNING: Ignored write to ECP register\n");
      break;
    default:
      fprintf (stderr,
               "ERROR: outportb() tried to write to an unsupported port (0x%x)\n",
               port);
      exit (1);
    }
#elif   defined __BEOS__
  st_ioport_t temp;

  temp.port = port;
  temp.data8 = byte;
  ioctl (misc_io_fd, 'w', &temp, 0);
#elif   defined AMIGA
  ULONG wait_mask;
  (void) port;                                  // warning remover

  misc_io_req->io_Length = 1;
  misc_io_req->io_Data = byte;
  misc_io_req->io_Command = CMD_WRITE;

/*
  SendIO ((struct IORequest *) misc_io_req);

  wait_mask = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F | 1L << misc_parport->mp_SigBit;
  if (Wait (wait_mask) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F))
    AbortIO ((struct IORequest *) misc_io_req);
  WaitIO ((struct IORequest *) misc_io_req);
*/

  if (DoIO ((struct IORequest *) misc_io_req))
    {
      fprintf (stderr, "ERROR: Could not communicate with parallel port:\n"
                       "       %s\n", misc_io_req->IOPar.io_Error);
      exit (1);
    }

#elif   defined _WIN32 || defined __CYGWIN__
  output_byte (port, byte);
#elif   defined __i386__
  i386_output_byte (port, byte);
#elif   defined HAVE_SYS_IO_H
  outb (byte, port);
#endif
}


void
outportw (unsigned short port, unsigned short word)
{
#ifdef  PPDEV
  int ppreg = port - ucon64.parport;
  unsigned char buf[2];

  // words are written in little endian format
  buf[0] = word;
  buf[1] = word >> 8;
  switch (ppreg)
    {
    case 3:                                     // EPP/ECP address
      if (!(misc_io_mode & IEEE1284_ADDR))      // IEEE1284_DATA is 0!
        {
          misc_io_mode |= IEEE1284_ADDR;
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      write (misc_io_fd, buf, 2);
      break;
    case 4:                                     // EPP/ECP data
      if (misc_io_mode & IEEE1284_ADDR)
        {
          misc_io_mode &= ~IEEE1284_ADDR;       // IEEE1284_DATA is 0
          ioctl (misc_io_fd, PPSETMODE, &misc_io_mode);
        }
      write (misc_io_fd, buf, 2);
      break;
    // the data, control and ECP registers should only be accessed in "8-bit mode"
    default:
      fprintf (stderr,
               "ERROR: outportw() tried to write to an unsupported port (0x%x)\n",
               port);
      exit (1);
    }
#elif   defined __BEOS__
  st_ioport_t temp;

  temp.port = port;
  temp.data16 = word;
  ioctl (misc_io_fd, 'w16', &temp, 0);
#elif   defined AMIGA
  ULONG wait_mask;
  (void) port;                                  // warning remover

  misc_io_req->io_Length = 2;
  misc_io_req->io_Data = word;
  misc_io_req->io_Command = CMD_WRITE;

/*
  SendIO ((struct IORequest *) misc_io_req);

  wait_mask = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F | 1L << misc_parport->mp_SigBit;
  if (Wait (wait_mask) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_F) )
    AbortIO ((struct IORequest *) misc_io_req);
  WaitIO ((struct IORequest *) misc_io_req);
*/

  if (DoIO ((struct IORequest *) misc_io_req))
    {
      fprintf (stderr, "ERROR: Could not communicate with parallel port:\n"
                       "       %s\n", misc_io_req->IOPar.io_Error);
      exit (1);
    }

#elif   defined _WIN32 || defined __CYGWIN__
  output_word (port, word);
#elif   defined __i386__
  i386_output_word (port, word);
#elif   defined HAVE_SYS_IO_H
  outw (word, port);
#endif
}


#define DETECT_MAX_CNT 1000
static int
misc_parport_probe (unsigned int port)
{
  int i = 0;

  outportb ((unsigned short) port, 0xaa);
  for (i = 0; i < DETECT_MAX_CNT; i++)
    if (inportb ((unsigned short) port) == 0xaa)
      break;

  if (i < DETECT_MAX_CNT)
    {
      outportb ((unsigned short) port, 0x55);
      for (i = 0; i < DETECT_MAX_CNT; i++)
        if (inportb ((unsigned short) port) == 0x55)
          break;
    }

  if (i >= DETECT_MAX_CNT)
    return 0;

  return 1;
}


int
misc_parport_open (int port)
{
#ifdef  PPDEV
  struct timeval t;
  int capabilities = 0, ucon64_parport, x;

  if (port == PARPORT_UNKNOWN)
    port = 0;

  misc_io_fd = open (ucon64.parport_dev, O_RDWR | O_NONBLOCK);
  if (misc_io_fd == -1)
    {
      fprintf (stderr, "ERROR: Could not open parallel port device (%s)\n"
                       "       Check if you have the required privileges\n",
                       ucon64.parport_dev);
      exit (1);
    }

  ioctl (misc_io_fd, PPEXCL);                   // disable sharing
  ioctl (misc_io_fd, PPCLAIM);
  t.tv_sec = 0;
  t.tv_usec = 500 * 1000;                       // set time-out to 500 milliseconds
  ioctl (misc_io_fd, PPSETTIME, &t);
/*
  ioctl (misc_io_fd, PPGETTIME, &t);
  printf ("Current time-out value: %ld microseconds\n", t.tv_usec);
*/

  ioctl (misc_io_fd, PPGETMODES, &capabilities);
//  printf ("Capabilities: %x\n", capabilities);

  if (ucon64.parport_mode == UCON64_EPP || ucon64.parport_mode == UCON64_ECP)
    if ((capabilities & (PARPORT_MODE_EPP | PARPORT_MODE_ECP)) == 0)
      printf ("WARNING: EPP or ECP mode was requested, but not available\n");

  if (capabilities & PARPORT_MODE_ECP)
    misc_io_mode = IEEE1284_MODE_ECP;
  else if (capabilities & PARPORT_MODE_EPP)
    misc_io_mode = IEEE1284_MODE_EPP;
  else
    misc_io_mode = IEEE1284_MODE_BYTE;
  misc_io_mode |= IEEE1284_DATA;                // default to EPP/ECP data reg
  ioctl (misc_io_fd, PPSETMODE, &misc_io_mode); //  (IEEE1284_DATA is 0...)

  x = PP_FASTREAD | PP_FASTWRITE;               // enable 16-bit transfers
  ioctl (misc_io_fd, PPSETFLAGS, &x);

  misc_io_direction = FORWARD;                  // set forward direction as default
  ioctl (misc_io_fd, PPDATADIR, &misc_io_direction);
#elif   defined __BEOS__
  misc_io_fd = open ("/dev/misc/ioport", O_RDWR | O_NONBLOCK);
  if (misc_io_fd == -1)
    {
      misc_io_fd = open ("/dev/misc/parnew", O_RDWR | O_NONBLOCK);
      if (misc_io_fd == -1)
        {
          fprintf (stderr, "ERROR: Could not open I/O port device (no driver)\n"
                           "       You can download the latest ioport driver from\n"
                           "       http://www.infernal.currantbun.com or http://ucon64.sourceforge.net\n");
          exit (1);
        }
      else
        {                                       // print warning, but continue
          printf ("WARNING: Support for the driver parnew is deprecated. Future versions of uCON64\n"
                  "         might not support this driver. You can download the latest ioport\n"
                  "         driver from http://www.infernal.currantbun.com or\n"
                  "         http://ucon64.sourceforge.net\n\n");
        }
    }
#elif   defined AMIGA
  int x;

  misc_parport = CreatePort (NULL, 0);
  if (misc_parport == NULL)
    {
      fprintf (stderr, "ERROR: Could not create the MsgPort\n");
      exit (1);
    }
  misc_io_req = CreateExtIO (misc_parport, sizeof (struct IOExtPar));
  if (misc_io_req == NULL)
    {
      fprintf (stderr, "ERROR: Could not create the I/O request structure\n");
      DeletePort (misc_parport);
      misc_io_req = NULL;
      exit (1);
    }

  // Is it possible to probe for the correct port?
  if (port == PARPORT_UNKNOWN)
    port = 0;

  x = OpenDevice (ucon64.parport_dev, port, (struct IORequest *) misc_io_req,
                  (ULONG) 0);
  if (x != 0)
    {
      fprintf (stderr, "ERROR: Could not open parallel port (%s, %x)\n",
               ucon64.parport_dev, port);
      DeleteExtIO ((struct IOExtPar *) misc_io_req);
      DeletePort (misc_parport);
      exit (1);
    }

  if (register_func (close_io_port) == -1)
    {
      // AbortIO ((struct IORequest *) misc_io_req); // should not be necessary with DoIO()
      CloseDevice ((struct IORequest *) misc_io_req);
      DeleteExtIO (misc_io_req);
      DeletePort (misc_parport);
      misc_io_req = NULL;
      fprintf (stderr, "ERROR: Could not register function with register_func()\n");
      exit (1);
    }
#elif   defined __FreeBSD__
  misc_io_fd = open ("/dev/io", O_RDWR);
  if (misc_io_fd == -1)
    {
      fprintf (stderr, "ERROR: Could not open I/O port device (/dev/io)\n"
                       "       (This program needs root privileges for the requested action)\n");
      exit (1);
    }
#endif

#if     defined PPDEV || defined __BEOS__ || defined __FreeBSD__
  if (register_func (close_io_port) == -1)
    {
      close (misc_io_fd);
      fprintf (stderr, "ERROR: Could not register function with register_func()\n");
      exit(1);
    }
#endif

#if     defined __linux__ && !defined __powerpc__ && !defined PPDEV
  /*
    Some code needs us to switch to the real uid and gid. However, other code
    needs access to I/O ports other than the standard printer port registers.
    We just do an iopl(3) and all code should be happy. Using iopl(3) enables
    users to run all code without being root (of course with the uCON64
    executable setuid root).
    Another reason to use iopl() and not ioperm() is that the former enables
    access to all I/O ports, while the latter enables access to ports up to
    0x3ff. For the standard parallel port hardware addresses this is not a
    problem. It *is* a problem for add-on parallel port cards which can be
    mapped to I/O addresses above 0x3ff.
  */
  if (iopl (3) == -1) // __NetBSD__ => (i386_iopl (3) != 0), add after feature request ;-)
    {
      fprintf (stderr, "ERROR: Could not set the I/O privilege level to 3\n"
                       "       (This program needs root privileges for the requested action)\n");
      exit (1);                                 // Don't return, if iopl() fails port access
    }                                           //  causes core dump
#endif // __linux__ && !__powerpc__ && !PPDEV

#if     (defined __i386__ || defined _WIN32) && !defined PPDEV

#if     defined _WIN32 || defined __CYGWIN__
  /*
    We support the I/O port drivers inpout32.dll and io.dll, because using them
    is way easier than using UserPort or GiveIO. The drivers are also more
    reliable and seem to enable access to all I/O ports (at least it's *very*
    easy to crash Windows XP ;-) The only downside to inpout32.dll is that it's
    almost two times slower than UserPort...
  */
  char fname[FILENAME_MAX];
  int driver_found = 0;

  sprintf (fname, "%s" FILE_SEPARATOR_S "%s", ucon64.configdir, "io.dll");
#if 0 // We must not do this for Cygwin or access() won't "find" the file
  change_mem (fname, strlen (fname), "/", 1, 0, 0, "\\", 1, 0);
#endif
  if (access (fname, F_OK) == 0)
    {
      io_driver = open_module (fname);

      IsDriverInstalled =
#ifdef  __cplusplus // this is really nice: gcc wants something else than g++...
                          (short int (WINAPI *) ())
#endif
                          get_symbol (io_driver, "IsDriverInstalled");
      if (IsDriverInstalled ())
        {
          driver_found = 1;
          printf ("Using %s\n", fname);

          PortIn =
#ifdef  __cplusplus
                   (char (WINAPI *) (short int))
#endif
                   get_symbol (io_driver, "PortIn");
          PortWordIn =
#ifdef  __cplusplus
                       (short int (WINAPI *) (short int))
#endif
                       get_symbol (io_driver, "PortWordIn");
          PortOut =
#ifdef  __cplusplus
                    (void (WINAPI *) (short int, char))
#endif
                    get_symbol (io_driver, "PortOut");
          PortWordOut =
#ifdef  __cplusplus
                        (void (WINAPI *) (short int, short int))
#endif
                        get_symbol (io_driver, "PortWordOut");
          input_byte = io_input_byte;
          input_word = io_input_word;
          output_byte = io_output_byte;
          output_word = io_output_word;
        }
    }

  if (!driver_found)
    {
      sprintf (fname, "%s" FILE_SEPARATOR_S "%s", ucon64.configdir, "inpout32.dll");
      if (access (fname, F_OK) == 0)
        {
          driver_found = 1;
          printf ("Using %s\n", fname);
          io_driver = open_module (fname);
          Inp32 =
#ifdef  __cplusplus
                  (unsigned char (__stdcall *) (unsigned short))
#endif
                  get_symbol (io_driver, "Inp32");
          Outp32 =
#ifdef  __cplusplus
                   (void (__stdcall *) (unsigned short, unsigned char))
#endif
                   get_symbol (io_driver, "Out32");
          // note that inport_word and output_word keep their default value...
          input_byte = inpout32_input_byte;
          output_byte = inpout32_output_byte;
        }
    }
#endif // _WIN32 || __CYGWIN__

  if (port == PARPORT_UNKNOWN)                   // no port specified or forced?
    {
      unsigned int parport_addresses[] = { 0x3bc, 0x378, 0x278 };
      int x, found = 0;

      for (x = 0; x < 3; x++)
        if ((found = misc_parport_probe (parport_addresses[x])) == 1)
          {
            port = parport_addresses[x];
            break;
          }

      if (found != 1)
        {
          fprintf (stderr, "ERROR: Could not find a parallel port on your system\n"
                           "       Try to specify it by hand\n\n");
          exit (1);
        }
    }
#endif // (__i386__ || _WIN32) && !PPDEV

#ifdef  PPDEV
  // the following two calls need a valid value for ucon64.parport
  ucon64_parport = ucon64.parport;
  ucon64.parport = port;
#endif
  outportb ((unsigned short) (port + PARPORT_CONTROL),
            (unsigned char) (inportb ((unsigned short) (port + PARPORT_CONTROL)) & 0x0f));
  // bit 4 = 0 -> IRQ disable for ACK, bit 5-7 unused
#ifdef  PPDEV
  ucon64.parport = ucon64_parport;
#endif

  return port;
}


int
misc_parport_close (int parport)
{
  (void) parport;
#if     defined PPDEV || defined __BEOS__ || defined __FreeBSD__ || defined AMIGA
  if (unregister_func (close_io_port) == 0)     // call func only if it can be removed!
    close_io_port ();                           //  (or else it will be called twice)
#elif   defined _WIN32 || defined __CYGWIN__
  input_byte = NULL;
  input_word = NULL;
  output_byte = NULL;
  output_word = NULL;
#endif
  return 0;
}


void
misc_parport_print_info (void)
{
#ifdef  PPDEV
  printf ("Using parallel port device: %s\n", ucon64.parport_dev);
#elif   defined AMIGA
  printf ("Using parallel port device: %s, port %d\n", ucon64.parport_dev, ucon64.parport);
#else
  printf ("Using I/O port base address: 0x%x\n", ucon64.parport);
#endif
}
#endif // PARALLEL
